import torch
from itertools import permutations

def ctmrg(Tin, d, Dcut, no_iter):
    
    #symmetrize
    T = torch.zeros_like(Tin)
    for i in permutations([0, 1, 2, 3], 4):
        T = T + Tin.permute(i)
    T = T/24. 
    
    lnZ = 0.0
    truncation_error = 0.0
    C = T.sum((0,1))
    E = T.sum(1)
    D = d
    for n in range(no_iter):
        A = torch.einsum('ab,eca,bdg,cdfh->efgh', (C, E, E, T)).contiguous().view(D*d, D*d)
        A = (A+A.t())/2.

        D_new = min(D*d, Dcut)
        U, S, V = torch.svd(A)
        truncation_error += S[D_new:].sum()/S.sum() 
        P = U[:, :D_new] # projection operator

        #S, U = torch.symeig(A, eigenvectors=True)
        #truncation_error += 1.-S[-D_new:].sum()/S.sum() 
        #P = U[:, -D_new:] # projection operator

        C = (P.t() @ A @ P) #(D, D)
        C = (C+C.t())/2.

        ET = torch.einsum('ldr,adbc->labrc', (E, T)).contiguous().view(D*d, d, D*d)
        #ET = torch.tensordot(E, T, dims=([1], [1]))
        #ET = ET.permute(0, 2, 3, 1, 4).contiguous().view(D*d, d, D*d)

        E = torch.einsum('li,ldr,rj->idj', (P, ET, P)) #(D_new, d, D_new)
        #E = ( P.t() @ ((ET.view(D*d*d, D*d)@P).view(D*d, d*D_new))).view(D_new,d,D_new)

        E = (E + E.permute(2, 1, 0))/2.

        D = D_new

        maxval = C.max()
        C = C/maxval 

        maxval = E.max()
        E = E/maxval 
    
    Z1 = torch.einsum('ab,bcd,fd,gha,hcij,fjk,lg,mil,mk', (C,E,C,E,T,E,C,E,C))
    #CEC = torch.einsum('da,ebd,ce->abc', (C,E,C)).view(1, D**2*d)
    #ETE = torch.einsum('abc,lbdr,mdn->almcrn',(E,T,E)).contiguous().view(D**2*d, D**2*d)
    #ETE = (ETE+ETE.t())/2.
    #Z1 = CEC@ETE@CEC.t()

    Z3 = torch.einsum('ab,bc,cd,da', (C,C,C,C))
    Z2 = torch.einsum('ab,bcd,de,fa,gcf,ge',(C,E,C,C,E,C))
    print ('Z1, Z2, Z3:', Z1.item(), Z2.item(), Z3.item())
    lnZ += torch.log(Z1.abs()) + torch.log(Z3.abs()) - 2.*torch.log(Z2.abs())

    return lnZ, truncation_error

if __name__=='__main__':
    torch.set_num_threads(4)
    K = torch.tensor([0.44])
    Dcut = 180
    n = 50

    #Boltzmann factor on a bond M=LR^T
    M = torch.stack([torch.cat([torch.exp(K),  torch.exp(-K)]), 
                     torch.cat([torch.exp(-K), torch.exp(K)])
                    ])
    U, S, V = torch.svd(M)
    L = U*torch.sqrt(S) 
    R = V*torch.sqrt(S)

    #            L 
    #            |
    # T =  R^{T}-o-L
    #            |
    #            R^{T}

    T = torch.einsum('ai,aj,ak,al->ijkl', (L, L, R, R)) 
    lnZ, error = ctmrg(T, 2, Dcut, n)
    print (lnZ.item(), error)
